
#include "convert.hpp"

extern "C"
{
#include "wad.h"
};

#include <cstring>
#include <vector>
extern "C" 
{

#include <time.h>
#include <stdlib.h>
static zip_fileinfo* newzipinfo()
{
    const time_t tim = time(NULL);
    struct tm* curtime = localtime(&tim);
    
    if(!curtime) return NULL;
    
    tm_zip ztime;
    zip_fileinfo* zinfo;
    
    
    if(! (zinfo = (zip_fileinfo*)malloc(sizeof(zip_fileinfo))) ) return NULL;
    
    ztime.tm_sec = (curtime->tm_sec >= 60 ? 0 : curtime->tm_sec);
    ztime.tm_min = curtime->tm_min;
    ztime.tm_hour = curtime->tm_hour;
    ztime.tm_mday = curtime->tm_mday;
    ztime.tm_mon = curtime->tm_mon;
    ztime.tm_year = curtime->tm_year+1900u;
    
    zinfo->tmz_date = ztime;
    zinfo->dosDate = zinfo->internal_fa = zinfo->external_fa = 0;
    
    return zinfo;
}



};

const unsigned int numMapLumps = 10;
static const char* mapslumps[numMapLumps]
{
    "THINGS",
    "LINEDEFS",
    "SIDEDEFS",
    "VERTEXES",
    "SEGS",
    "SSECTORS",
    "NODES",
    "SECTORS",
    "REJECT",
    "BLOCKMAP"
};


//Convert function
//filename = the wad file
//savename = the name of the new archive file to be created where the lumps will be stored
//method = compression methods for the files inside the archive
//level = compression level for DEFLATE. ignored if the chosen compression method is other than deflate.
//omsg = ostream object where verbose output will be feed
bool convert2pk3(const char* filename, const char* savename, int method, int level, std::ostream& omsg)
{
    using namespace std::literals;
    zipFile nzip;
    zip_fileinfo* zinfo;
    
    std::vector<unsigned int> soclumps;
    std::vector<unsigned int> lualumps;
    std::vector<unsigned int> musiclumps;
    std::vector<unsigned int> soundlumps;
    std::vector<unsigned int> maplumps;
    std::vector<unsigned int> otherlumps;
    
    
    omsg << "Loading wad " << filename << "..." << std::endl;
    
    wad_t* wad;
    if(! (wad = W_loadwad(filename)) )
    {
        omsg << "Error on loading wad: " << W_geterrormsg() << std::endl;
        return false;
    }
    
    //Convert backslash characters to plus sign.
    for(unsigned int i=0;i<wad->numlumps;i++)
    {
        char* l = (wad->lumps+i)->name;
        while( (l = std::strchr(l,'\\') ) != NULL) *l = '+';
    }
    
    omsg << "Creating archive file " << savename << "..." << std::endl;
    if(! (nzip = zipOpen(savename,APPEND_STATUS_CREATE)) )
    {
        omsg << "could not create archive file " << savename << std::endl;
        return false;
    }
    
    if(! (zinfo = newzipinfo()) ) 
    {
        omsg << "could not retrieve a new zip info" << std::endl;
        return false;
    }
    //scan all the wads
    unsigned int i=0;
    while(i<wad->numlumps)
    {
        lump_t* lump = &wad->lumps[i];
        
        //is it a lump marker?
        if(!std::strncmp(lump->name,"FF_START",8) || !std::strncmp(lump->name, "F_START",7))
        {
            zipOpenNewFileInZip(nzip,"Flats/",zinfo,NULL,0,NULL,0,NULL,method,level);
            zipCloseFileInZip(nzip);
        
            for(++i; i<wad->numlumps;i++)
            {
                lump_t *lmp = &wad->lumps[i];
                if(!std::strncmp(lmp->name,"FF_END",6) || !std::strncmp(lmp->name,"F_END",5))
                    break;
                
                    
                zipOpenNewFileInZip(nzip,("Flats/"s+lmp->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
                zipWriteInFileInZip(nzip,lmp->data,lmp->size);
                zipCloseFileInZip(nzip);
            }
        }
        else if(!std::strncmp(lump->name,"TX_START",8))
        {
            zipOpenNewFileInZip(nzip,"Textures/",zinfo,NULL,0,NULL,0,NULL,method,level);
            zipCloseFileInZip(nzip);
        
            for(++i; i<wad->numlumps;i++)
            {
                lump_t *lmp = &wad->lumps[i];
                if(!std::strncmp(lmp->name,"TX_END",6))
                    break;
                    
                zipOpenNewFileInZip(nzip,("Textures/"s+lmp->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
                zipWriteInFileInZip(nzip,lmp->data,lmp->size);
                zipCloseFileInZip(nzip);
            }
        }
        else if(!std::strncmp(lump->name,"S_START",7) || !std::strncmp(lump->name,"SS_START",8))
        {
            zipOpenNewFileInZip(nzip,"Sprites/",zinfo,NULL,0,NULL,0,NULL,method,level);
            zipCloseFileInZip(nzip);
        
            for(++i; i<wad->numlumps;i++)
            {
                lump_t *lmp = &wad->lumps[i];
                if(!std::strncmp(lmp->name,"S_END",5) || !std::strncmp(lmp->name,"SS_END",6))
                    break;
                
                zipOpenNewFileInZip(nzip,("Sprites/"s+lmp->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
                zipWriteInFileInZip(nzip,lmp->data,lmp->size);
                zipCloseFileInZip(nzip);
            }
        }
        else if(!std::strncmp(lump->name,"P_START",7))
        {
            zipOpenNewFileInZip(nzip,"Patches/",zinfo,NULL,0,NULL,0,NULL,method,level);
            zipCloseFileInZip(nzip);
        
            for(++i; i<wad->numlumps;i++)
            {
                lump_t *lmp = &wad->lumps[i];
                if(!std::strncmp(lmp->name,"P_END",5))
                    break;
                
                zipOpenNewFileInZip(nzip,("Patches/"s+lmp->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
                zipWriteInFileInZip(nzip,lmp->data,lmp->size);
                zipCloseFileInZip(nzip);
            }
        } else if(!std::strncmp(lump->name,"SOC_",4) || !std::strncmp(lump->name,"MAINCFG",7) || !std::strncmp(lump->name,"OBJCTCFG",8))
        {
            soclumps.push_back(i);
        }
        else if(!std::strncmp(lump->name,"LUA_",4))
        {
            lualumps.push_back(i);
        }
        else if(!std::strncmp(lump->name,"DS",2))
        {
            soundlumps.push_back(i);
        }
        else if(!std::strncmp(lump->name,"O_",2) || !std::strncmp(lump->name,"D_",2))
        {
            musiclumps.push_back(i);
        }
        else if(!std::strncmp(lump->name,"MAP",3))
        {
            maplumps.push_back(i);
        }
        else
        {
            bool isMaplump=false;
            //Check for map lumps
            for(unsigned int j=0;j<numMapLumps;j++)
            {
                if(!std::strncmp(lump->name,mapslumps[j],strlen(mapslumps[j])-1))
                {
                    maplumps.push_back(i);
                    
                    isMaplump=true;
                    break;
                }
            }
            
            if(!isMaplump)
            {
                otherlumps.push_back(i);
            }
        }
        
        i++;
    }
    
    if(!soclumps.empty())
    {
        zipOpenNewFileInZip(nzip,"SOC/",zinfo,NULL,0,NULL,0,NULL,method,level);
        zipCloseFileInZip(nzip);
    
        for(unsigned int j : soclumps)
        {
            lump_t* lump = wad->lumps+j;
            
            zipOpenNewFileInZip(nzip,("SOC/"s+lump->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
            zipWriteInFileInZip(nzip,lump->data,lump->size);
            zipCloseFileInZip(nzip);
        }
    }
    
    if(!lualumps.empty())
    {
        zipOpenNewFileInZip(nzip,"LUA/",zinfo,NULL,0,NULL,0,NULL,method,level);
        zipCloseFileInZip(nzip);
        
        for(unsigned int j : lualumps)
        {
            lump_t* lump = wad->lumps+j;
            
            zipOpenNewFileInZip(nzip,("LUA/"s+lump->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
            zipWriteInFileInZip(nzip,lump->data,lump->size);
            zipCloseFileInZip(nzip);
        }
    }
    
    if(!musiclumps.empty())
    {
        zipOpenNewFileInZip(nzip,"Music/",zinfo,NULL,0,NULL,0,NULL,method,level);
        zipCloseFileInZip(nzip);
        
        for(unsigned int j : musiclumps)
        {
            lump_t* lump = wad->lumps+j;
            
            zipOpenNewFileInZip(nzip,("Music/"s+lump->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
            zipWriteInFileInZip(nzip,lump->data,lump->size);
            zipCloseFileInZip(nzip);
        }
    }
    
    if(!soundlumps.empty())
    {
        zipOpenNewFileInZip(nzip,"Sounds/",zinfo,NULL,0,NULL,0,NULL,method,level);
        zipCloseFileInZip(nzip);
        
        for(unsigned int j : soundlumps)
        {
            lump_t* lump = wad->lumps+j;
            
            zipOpenNewFileInZip(nzip,("Sounds/"s+lump->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
            zipWriteInFileInZip(nzip,lump->data,lump->size);
            zipCloseFileInZip(nzip);
        }
    }
    
    if(!maplumps.empty())
    {
        zipOpenNewFileInZip(nzip,"Maps/",zinfo,NULL,0,NULL,0,NULL,method,level);
        zipCloseFileInZip(nzip);
        
        for(unsigned int j : maplumps)
        {
            lump_t* lump = wad->lumps+j;
            
            zipOpenNewFileInZip(nzip,("Maps/"s+lump->name).c_str(),zinfo,NULL,0,NULL,0,NULL,method,level);
            zipWriteInFileInZip(nzip,lump->data,lump->size);
            zipCloseFileInZip(nzip);
        }
    }
    
    if(!otherlumps.empty())
    {
        for(unsigned int j : otherlumps)
        {
            lump_t* lump = wad->lumps+j;
            
            zipOpenNewFileInZip(nzip,lump->name,zinfo,NULL,0,NULL,0,NULL,method,level);
            zipWriteInFileInZip(nzip,lump->data,lump->size);
            zipCloseFileInZip(nzip);
        }
    }
        
    omsg << "closing wad and archive" << std::endl;
    W_closewad(wad);
    omsg << "closed wad" << std::endl;
    zipClose(nzip,NULL);
    
    omsg << "Done writing output file " << savename << std::endl;
    
    return true;
}